import tempfile

import cv2
import geojson
import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.affinity as shaff
from box import Box
from osgeo import gdal
from scipy import ndimage

from . import gdal_datasets as gds,Folder
from . import logger
from .common.utils import run_cmd
from .pathify import get_name

class Vectorizer(object):
	def __init__(self,ds,array=None,no_data_value=-9999,**history):
		self.ds=gds.DataSet(ds,array=array)
		self.history=Box(history,default_box=True)
		self.no_data_value=no_data_value
		self._data=np.where(self.ds.array==no_data_value,np.nan,self.ds.array)
		logger.debug('_data.shape: %s',self._data.shape)
		# self.history.original_data=self._data
		single_pixel_area=abs(self.ds.transform[1]*self.ds.transform[-1])/1e4
		self.area=np.count_nonzero(~np.isnan(self._data))*single_pixel_area
		logger.debug('area (Ha):%s',self.area)
		self.__geo_info=self.ds.geo_info
		self._geometry=None
		self.history.original.geo_info=self.ds.geo_info
	
	@property
	def geometry(self):
		if self._geometry is None:
			self._geometry=Vectorizer(self.ds.path,
			                          no_data_value=self.no_data_value)\
				.vectorize(1)\
				.simplify()\
				.history.vectors.geometry.cascaded_union
		return self._geometry
	
	@property
	def data(self) -> np.ndarray:
		return self._data
	
	@property
	def geo_info(self):
		return self.__geo_info
	
	@property
	def raster_area(self):
		single_pixel_area=abs(self.ds.transform[1]*self.ds.transform[-1])/1e4
		return np.count_nonzero(~np.isnan(self._data))*single_pixel_area
	
	@staticmethod
	def get_resized_array(array,scale_factor):
		resized_data=cv2.resize(array,tuple(np.round(np.array(array.shape)*scale_factor).astype(int))[::-1],interpolation=cv2.INTER_CUBIC)
		return resized_data
	
	@logger.trace()
	def resize(self,scale_factor):
		# Plotter(self.data)
		self.history.scale_factor=scale_factor
		resized_data=self.get_resized_array(self.data,scale_factor=scale_factor)
		logger.debug('resized_data.shape: %s',resized_data.shape)
		# self.history.resized_data=resized_data
		self._data=resized_data
		transform,projection=self.geo_info
		transform=list(transform)
		transform[1]/=scale_factor
		transform[-1]/=scale_factor
		self.__geo_info=transform,projection
		logger.debug('new_geo_info: %s',self.__geo_info)
		
		return self
	
	@logger.trace()
	def smooth(self,sigma=1):
		logger.debug('self.data.shape: %s',self.data.shape)
		self._data=ndimage.gaussian_filter(self.data,sigma=sigma)
		# Plotter().add_images(self.data,res).plot()
		
		return self
	
	@staticmethod
	def get_digitized(array,bins):
		if isinstance(bins,(int,np.integer)):
			bin_edges=np.histogram_bin_edges(array[~np.isnan(array)],bins)
		else:
			bin_edges=bins
		logger.debug('bin_edges: %s',bin_edges)
		digitized_data=np.digitize(array,bin_edges)
		return digitized_data,bin_edges
	
	@staticmethod
	def get_georeferenced_vectors(vectors: gpd.GeoDataFrame,transform,raster_shape):
		vectors.geometry=vectors.geometry.map(
			lambda g:shaff.scale(g,xfact=transform[1],yfact=abs(transform[5]),origin=(0,0))
		)
		vectors.geometry=vectors.geometry.map(
			lambda g:shaff.translate(g,transform[0],transform[3]+transform[5]*raster_shape[0]))
		return vectors
	
	@staticmethod
	def get_vectorized(array,bins,geo_info,**kwargs):
		# old_array=array
		# if isinstance(bins,(int,np.integer)) and bins<0:
		
		array,bin_edges=Vectorizer.get_digitized(array,bins)
		# Plotter(old_array,array)
		vectors=list()
		unique_values=np.unique(array)
		potrace_folder=Folder(tempfile.mkdtemp())
		for unique_value_i,unique_value in enumerate(unique_values):
			# logger.debug('unique_value: %s',unique_value)
			tmp_folder=potrace_folder[f'unique_value_{unique_value}']
			if unique_value==len(bin_edges):
				# logger.info('Continuing')
				continue
			value_mask=np.where(array==unique_value,0,255)
			mask_ds=gds.DataSet.from_array(value_mask,geo_info)
			bmp_filepath=mask_ds.to_file(filepath=tmp_folder[f'mask_{unique_value}.bmp'],
			                             driver_name='BMP',no_data_value=0,dtype=gdal.GDT_Byte)
			vector_path=tmp_folder[f"out_{unique_value}.geojson"]
			cmd_kwargs=" ".join([f'--{k} {v}' for k,v in kwargs.items()])
			run_cmd(f'potrace -o {vector_path} -b geojson {bmp_filepath} '+cmd_kwargs)
			df: gpd.GeoDataFrame=gpd.GeoDataFrame.from_file(vector_path)
			df['DN']=np.mean(bin_edges[unique_value_i:unique_value_i+2])
			# logger.debug('df:\n%s',df)
			vectors.append(df)
		potrace_folder.delete()
		
		combined_vectors=Vectorizer.get_georeferenced_vectors(
			vectors=gpd.GeoDataFrame(pd.concat(vectors,ignore_index=True),crs=3857),
			transform=geo_info[0],
			raster_shape=array.shape)
		return combined_vectors
	
	@logger.trace()
	def vectorize(self,bins,**kwargs):
		# Plotter().add_images(self.data).plot()
		vectors: gpd.GeoDataFrame=self.get_vectorized(self.data,
		                                              bins=bins,
		                                              geo_info=self.geo_info,
		                                              **kwargs)
		self.history.vectors=vectors
		return self
	
	# @staticmethod
	# def _simplify_polygon(s: gpd.GeoSeries,decimals=3,area_threshold=0):
	# 	simplifier=lambda linering:ThePoly(simplify_coords(np.round(ThePoly(linering)
	# 	                                                            .xy,decimals=decimals)
	# 	                                                   .copy(order='C'),1)).geometry
	#
	# 	g: shg.Polygon=s.geometry
	#
	# 	g_exterior=simplifier(g.exterior).exterior.coords
	# 	g_interiors=list()
	# 	# has_interior=False
	# 	for interior in g.interiors:
	# 		interior_geom=simplifier(interior)
	# 		if interior_geom.area<area_threshold:
	# 			logger.debug('interior_geom.area: %s',interior_geom.area)
	# 			continue
	# 		# self.history.interiors_lookup_data.append(dict(
	# 		# 	parent_index=s.name,geometry=interior_geom
	# 		# ))
	# 		# has_interior=True
	# 		g_interiors.append(interior_geom.exterior.coords)
	# 	out_g=shg.Polygon(g_exterior,g_interiors)
	#
	# 	# if has_interior:
	# 	# 	out_g_no_holes=shg.Polygon(g_exterior)
	# 	# 	logger.debug('out_g_no_holes: %s',out_g_no_holes)
	# 	# 	out_g_with_holes=shg.Polygon(g_exterior,g_interiors)
	# 	# 	logger.debug('out_g_with_holes: %s',out_g_with_holes)
	# 	#
	# 	# 	ThePoly(g).plot('input_geometry',c='green')
	# 	# 	ThePoly(out_g).plot('output_geometry',c='red')
	# 	# 	plt.show()
	# 	# g=ThePoly(simplify_coords(ThePoly(g).xy.copy(order='C'))).geometry
	# 	return out_g
	
	# @staticmethod
	# def _make_difference(g1,g2,max_tries=100):
	# 	try_count=0
	# 	while try_count<max_tries:
	# 		try_count+=1
	# 		try:
	# 			res=g1.difference(g2)
	# 			# logger.debug('res: %s',res)
	# 			return res.iloc[0]
	# 		except shapely.errors.TopologicalError:
	# 			if not isinstance(g2,shg.MultiPolygon): raise ValueError
	# 			g2: shg.MultiPolygon=g2
	# 			out_polys=list()
	# 			for poly in g2:
	# 				poly: shg.Polygon=poly
	# 				out_poly=ThePoly(ThePoly(poly).xy+np.random.rand()*1e-6).geometry
	# 				for interior in poly.interiors:
	# 					logger.debug('interior: %s',interior)
	# 					out_poly=out_poly.difference(shg.Polygon(interior))
	#
	# 				out_polys.append(out_poly)
	# 			out_multipolygon=shg.MultiPolygon(out_polys)
	# 			return out_multipolygon
	# 	# logger.debug('type(g2): %s',type(g2))
	# 	# for g2.exterior
	# 	# g2=shaff.scale(g2,xfact=1+np.random.rand()*1e-3*(try_count+1),yfact=1+np.random.rand()*1e-3*(try_count+1))
	# 	raise NotImplementedError
	
	# @staticmethod
	# def get_united(vectors: gpd.GeoDataFrame,geo_info):
	# 	vector_path=Vectorizer.get_saved(vectors,tempfile.mktemp('.geojson'))
	# 	transform=geo_info[0]
	# 	# vector_path=r'/home/lgblkb/PycharmProjects/imagination/tests/vector_results/ndvi_lgblkb_test_1_13.geojson'
	# 	# this_folder['rasters'].get_filepath('raster_result',ext='.tiff',iterated=True)
	# 	raster_path=rasterize(vector_path,tempfile.mktemp('.tiff'),burn_value=1,
	# 	                      x_res=transform[1],y_res=abs(transform[-1]),NoData_value=0)
	# 	logger.debug('raster_path: %s',raster_path)
	# 	vectors=Vectorizer(raster_path,no_data_value=0)\
	# 		.vectorize(bins=1).history.vectors
	# 	#get_vectorized(array=gds.DataSet(raster_path).array,bins=1,geo_info=geo_info)
	# 	os.remove(vector_path)
	# 	os.remove(raster_path)
	# 	logger.debug('vectors:\n%s',vectors)
	# 	return vectors
	
	# @staticmethod
	# def get_simplified(vectors: gpd.GeoDataFrame,geo_info,scale_factor,biggest_n=200,**kwargs):
	# 	vectors.geometry=vectors.apply(partial(Vectorizer._simplify_polygon,**kwargs),axis=1)
	#
	# 	# vectors=vectors.copy()
	# 	# vectors['area']=vectors.area
	# 	# total_area=vectors.area.sum()
	# 	# logger.debug('total_area: %s',total_area)
	# 	# vectors=vectors.sort_values(by=['area'],ascending=False).drop(columns=['area'])
	# 	# target_vectors: gpd.GeoDataFrame=vectors.iloc[:biggest_n,:].copy()
	# 	# logger.debug('target_vectors:\n%s',target_vectors)
	# 	# targets_area=target_vectors.area.sum()
	# 	# logger.debug('targets_area: %s',targets_area)
	# 	# delta_area=total_area-targets_area
	# 	# logger.debug('delta_area: %s',delta_area)
	#
	# 	# target_vectors.geometry=target_vectors.apply(partial(
	# 	# 	Vectorizer._simplify_polygon,area_threshold=target_vectors.area.min(),**kwargs),axis=1)
	# 	# smaller_vectors: gpd.GeoDataFrame=vectors.iloc[biggest_n:,:].copy()
	# 	# logger.debug('smaller_vectors:\n%s',smaller_vectors)
	# 	logger.debug('vectors:\n%s',vectors)
	# 	return vectors
	
	@logger.trace()
	def simplify(self,percent=10):
		temp_vector_path=self.save(tempfile.mktemp('.geojson'))
		output_path=tempfile.mktemp('.geojson')
		logger.debug('output_path: %s',output_path)
		run_cmd(f'mapshaper {temp_vector_path} -simplify {percent}% -o {output_path}',check=False)
		self.history.vectors=gpd.read_file(output_path,crs=3857)
		logger.debug('self.history.vectors:\n%s',self.history.vectors)
		
		# self.history.vectors=self.get_simplified(vectors=self.history.vectors,
		#                                          geo_info=self.geo_info,
		#                                          scale_factor=self.history.get('scale_factor',1),
		#                                          **kwargs)
		return self
	
	# @logger.trace()
	# def simplify_old(self,biggest_n=200,**kwargs):
	#
	# 	vectors: gpd.GeoDataFrame=self.history.combined_vectors
	# 	# self.history.interiors_lookup_data=list()
	# 	vectors.geometry=vectors.apply(partial(self._simplify_polygon,**kwargs),axis=1)
	# 	# interiors_lookup_table=ilt=gpd.GeoDataFrame(self.history.interiors_lookup_data)
	# 	# logger.debug('interiors_lookup_table:\n%s',interiors_lookup_table)
	# 	vectors['area']=vectors.area
	# 	vectors.sort_values(by=['area'],ascending=False,inplace=True)
	# 	target_vectors=vectors.iloc[:biggest_n,:].copy()
	# 	targets_area=target_vectors.area.sum()
	# 	logger.debug('targets_area: %s',targets_area)
	#
	# 	logger.debug('target_vectors:\n%s',target_vectors)
	#
	# 	# target_interior_geoms=ilt[ilt.parent_index.isin(target_vectors.index)].copy()
	# 	# target_interior_geoms=target_interior_geoms[target_interior_geoms.area<target_vectors.area.min()]
	# 	# logger.debug('target_interior_geoms:\n%s',target_interior_geoms)
	# 	# target_interior_geoms
	# 	# target_vectors.geometry=target_vectors.geometry.map(partial(self._simplify_polygon,**kwargs))
	# 	target_vectors.geometry=target_vectors.apply(partial(self._simplify_polygon,
	# 	                                                     area_threshold=target_vectors.area.min(),
	# 	                                                     **kwargs),axis=1)
	#
	# 	# ds_geom:gpd.GeoSeries=Vectorizer(self.ds,no_data_value=self.no_data_value)\
	# 	# 	.digitize(1).vectorize().history.combined_vectors.geometry
	#
	# 	# simplify_geom=lambda g:ThePoly(np.round(ThePoly(g).xy,decimals=decimals)).geometry
	# 	# vectors.geometry=vectors.geometry.map(simplify_geom)
	# 	# logger.debug('vectors.area:\n%s',vectors.area)
	# 	# groups=list()
	# 	# for dn,group in vectors.groupby('dn'):
	# 	# 	group: gpd.GeoDataFrame=group
	# 	# 	plt.hist(group.area,bins=100)
	# 	# 	plt.show()
	#
	# 	self.history.combined_vectors=target_vectors  #=vectors[vectors.area>min_area]
	# 	logger.debug('vectors:\n%s',self.history.combined_vectors)
	# 	return self
	
	# @logger.trace()
	# def differentiate(self):
	# 	vectors: gpd.GeoDataFrame=self.history.combined_vectors
	# 	vectors['area']=vectors.area
	# 	vectors=vectors.sort_values('area',ascending=True).drop(columns=['area']).reset_index(drop=True)
	# 	# logger.debug('vectors:\n%s',vectors)
	# 	polys=list()
	# 	for first_item in range(vectors.shape[0]):
	# 		logger.debug('item: %s',first_item)
	# 		previous_polys: gpd.GeoDataFrame=vectors.iloc[:first_item,:]
	# 		current_poly: gpd.GeoDataFrame=vectors.iloc[[first_item],:]
	# 		# next_polys=vectors.iloc[first_item+1:,:]
	# 		# logger.debug('current_poly:\n%s',current_poly)
	# 		# logger.debug('next_polys:\n%s',next_polys)
	# 		# if next_polys.empty:
	# 		# 	intersection=current_poly
	# 		# else:
	# 		# 	intersection=gpd.overlay(current_poly,next_polys,how='intersection')
	# 		# logger.debug('intersection:\n%s',intersection)
	# 		# logger.debug('current_poly:\n%s',current_poly)
	# 		# logger.debug('previous_polys:\n%s',previous_polys)
	# 		# logger.debug('previous_polys.cascaded_union:\n%s',previous_polys.cascaded_union)
	# 		# continue
	# 		difference=self._make_difference(current_poly,
	# 		                                 previous_polys.cascaded_union)  #current_poly.difference(previous_polys.cascaded_union)
	# 		# logger.debug('difference:\n%s',difference)
	#
	# 		# logger.debug('difference:\n%s',difference)
	# 		# logger.debug('type(difference): %s',type(difference))
	# 		if difference is None: continue
	# 		current_poly.geometry=[difference]
	# 		# logger.debug('current_poly:\n%s',current_poly)
	# 		# logger.debug('difference:\n%s',difference)
	# 		# if isinstance(difference,gpd.GeoSeries): difference=gpd.GeoDataFrame(difference).T
	#
	# 		# if previous_polys.empty:
	# 		# 	difference=current_poly
	# 		# else:
	# 		# 	difference=gpd.overlay(current_poly,previous_polys,how='difference')\
	# 		# 		.drop(columns=['dn_2'],errors='ignore').rename(columns=dict(dn_1='dn'))
	# 		# logger.debug('difference:\n%s',difference)
	# 		# # logger.debug('type(difference):\n%s',type(difference))
	# 		# logger.debug('difference.empty: %s',difference.empty)
	# 		# if difference.empty: continue
	# 		polys.append(current_poly.explode())
	# 	# return self
	# 	# polys.append(vectors.iloc[[-1],:])  # add the last smallest poly
	# 	resultant_vectors=gpd.GeoDataFrame(pd.concat(polys),crs=3857).reset_index(drop=True)
	# 	logger.debug('resultant_vectors:\n%s',resultant_vectors)
	# 	self.history.combined_vectors=resultant_vectors
	# 	return self
	
	@staticmethod
	def get_saved(vectors: gpd.GeoDataFrame,vector_path):
		if 'DN' in vectors.columns: property_getter=lambda s:dict(DN=np.round(s.DN,decimals=5))
		else: property_getter=lambda s:None
		
		features=vectors.apply(lambda s:geojson.Feature(geometry=s.geometry,properties=property_getter(s)),axis=1)
		# logger.debug('features:\n%s',features)
		data=geojson.FeatureCollection(features.tolist(),name=get_name(vector_path),
		                               crs=dict(type="name",
		                                        properties=dict(name="urn:ogc:def:crs:EPSG::3857")),)
		with open(vector_path,'w') as fh:
			geojson.dump(data,fh)
		logger.debug('vector_path: %s',vector_path)
		return vector_path
	
	@logger.trace()
	def save(self,vector_path):
		return self.get_saved(self.history.vectors,vector_path)



import functools
import glob
import os
import shutil
import zipfile
from datetime import datetime
from pathlib import Path, PosixPath
from typing import Callable

from box import Box
from checksumdir import dirhash

from . import logger

get_name = lambda some_path: os.path.splitext(os.path.basename(some_path))[0]


def get_parent_dir(some_path, depth=1):
    for i in range(depth):
        some_path = os.path.dirname(os.path.abspath(some_path))
    return some_path


def create_path(*paths, stop_depth=0):
    path = os.path.join(*paths)
    os.makedirs(get_parent_dir(path, stop_depth), exist_ok=True)
    return path


def get_existing_path(paths, safe=False):
    for p in paths:
        if os.path.exists(p): return p
    if not safe:
        raise FileNotFoundError('Could not find any existing path.')


def get_splitted(path):
    folders = []
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)
            break
    folders.reverse()
    return folders


def create_zipfile(dir_to_zip, savepath=''):
    """Create a zip file from all the files under 'dir_to_zip'.

    The output zip file will be saved to savepath.
    If savepath ends with '.zip', then the output zip file will be
    saved AS 'savepath'. Necessary tree subdirectories are created automatically.
    Else, savepath is assumed to be a directory path,
    hence the output zip file will be saved TO 'savepath'
    directory. Necessary tree subdirectories are created automatically.

    :return absolute savepath
    """
    save_cwd = os.getcwd()
    dir_to_zip = os.path.abspath(dir_to_zip)
    if dir_to_zip in os.path.split(savepath)[0]: raise ValueError(
        'To avoid recursion), resultant "savepath" should not be located inside "dir_to_zip"',
        dict(dir_to_zip=dir_to_zip, savepath=savepath))
    parent_dir, dir_name = os.path.split(dir_to_zip)
    os.chdir(parent_dir)
    if savepath:
        if savepath.endswith('.zip'):
            create_path(savepath, stop_depth=1)
        else:
            create_path(savepath, stop_depth=0)
            savepath = os.path.join(savepath, dir_name + '.zip')
    else:
        savepath = dir_to_zip + '.zip'

    pwd_length = len(os.getcwd())
    with zipfile.ZipFile(savepath, "w", compression=zipfile.ZIP_DEFLATED) as zf:
        for dirpath, dirnames, filenames in os.walk(dir_to_zip):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, path[pwd_length + 1:])

            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, path[pwd_length + 1:])

    os.chdir(save_cwd)
    return os.path.abspath(savepath)


def generate_name_portions(*args, __kv_delim__='-', **kwargs):
    arg_parts = [str(x) for x in args]
    kwarg_parts = [f"{k}{__kv_delim__}{v}" for k, v in kwargs.items() if v is not None]
    return arg_parts + kwarg_parts


def _get_name_parts_from_depth(path, include_depth):
    if type(include_depth) is Callable:
        parent_parts = include_depth(get_splitted(path))
    elif type(include_depth) is int:
        parent_parts = get_splitted(path)[-include_depth:]
    elif type(include_depth) in [tuple, list]:
        parent_parts = [get_splitted(path)[-x] for x in include_depth]
    elif type(include_depth) is slice:
        parent_parts = get_splitted(path)[-include_depth]
    else:
        raise ValueError(f'include_depth={include_depth}')
    return parent_parts


class Folder(os.PathLike):

    def __init__(self, path='', reactive=True, assert_exists=False):
        if isinstance(path, PosixPath):
            path = str(path)
        path = os.path.abspath(os.path.normpath(path))
        if self._is_file(path): path = get_parent_dir(path)
        if not self._exists(path):
            # possible_extension=os.path.splitext(path)[-1]
            # if possible_extension:
            # 	logger.warning('Provided folder path (which does not exist) appears to have an extension %s',possible_extension)
            # 	pass
            if assert_exists:
                raise AssertionError('Path should already exist.', dict(path=path))
            elif reactive:
                create_path(path)

        self.path = path
        self._class = functools.partial(self.__class__)
        self._inherit_class = self._class
        self.reactive = reactive

    @property
    def name(self) -> str:
        return os.path.split(self.path)[-1]

    def __str__(self):
        # return f"{self.__class__.__name__}(r'{self.path}')"
        return self.path

    def get_filepath(self, *name_args, ext='', delim='_', include_depth=None, datetime_loc_index=None, iterated=False,
                     **name_kwargs):
        parts = list()
        if include_depth: parts.extend(_get_name_parts_from_depth(self.path, include_depth))
        parts.extend(generate_name_portions(*name_args, **name_kwargs))
        if datetime_loc_index is not None: parts.insert(datetime_loc_index,
                                                        datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        assert parts, 'Nothing is provided to create filepath.'

        # output_path = os.path.join(self.path, delim.join(map(str, parts)).replace(' ', delim) + ext)
        output_path = os.path.join(self.path, delim.join(map(str, parts)) + ext)
        if iterated:
            return create_iterated_path(output_path)
        else:
            return output_path

    def _exists(self, path):
        return os.path.exists(path)

    def _is_file(self, path):
        return os.path.isfile(path)

    def _create_dir(self, dirpath, **kwargs):
        return self._inherit_class(dirpath, **kwargs)

    def create(self, *name_args, delim='_', datetime_loc_index=None, iterated=False, reactive=None, **name_kwargs):
        """
        :param name_args: args that are converted to name portions.
        :param delim: delimiter to use between the name string portions
        :param datetime_loc_index: index location of datetime in the resultant list of name portions.
        :param iterated: if iterated==True, method it will first check for existence of the folder. If exists, it creates a folder with new name,
        which is generated by incrementing the folder name.
        :param reactive: if not provided, default value is used. If provided, new folder will be created with specified value.
        :param name_kwargs: kwargs that are converted to name portions.
        :return: new instance of the class.
        :rtype: Folder
        """
        new_dirpath = self.get_filepath(*name_args, ext='', delim=delim, include_depth=None,
                                        datetime_loc_index=datetime_loc_index, iterated=iterated, **name_kwargs)
        # new_folder=self._create_dir(new_dirpath,**({} if reactive is None else dict(reactive=reactive)))
        new_folder = self._create_dir(new_dirpath, **dict(reactive=self.reactive if reactive is None else reactive))
        return new_folder

    def delete(self):
        # assert self.reactive,f'Folder {self.path} is not reactive'
        shutil.rmtree(self.path, ignore_errors=True)

    @logger.trace(skimpy=True)
    def clear(self):
        # assert self.reactive,f'Folder {self.path} is not reactive'
        self.delete()
        # try:
        #
        # except FileNotFoundError:
        # 	# lu.simple_logger.warning('FileNotFoundError when trying to clear up the results folder. Passing on.')
        # 	pass
        create_path(self.path)
        return self

    def glob_delete(self, *patterns, recursive=True):
        glob_search_results = self.glob_search(*patterns, recursive=recursive)
        for glob_search_result in glob_search_results:
            if os.path.isdir(glob_search_result):
                shutil.rmtree(glob_search_result)
            else:
                os.remove(glob_search_result)
        return self

    @property
    def children(self):
        return self.glob_search('*')

    def glob_search(self, *patterns, recursive=True):
        return glob.glob(self.get_filepath(*patterns), recursive=recursive)

    @logger.trace(skimpy=True)
    def zip(self, zip_filepath='', forced=False):
        zip_filepath = zip_filepath or self.parent().get_filepath(self.name, ext='.zip')
        if len(get_splitted(zip_filepath)) == 1: zip_filepath = self.parent().get_filepath(zip_filepath)
        if not zip_filepath.endswith('.zip'): zip_filepath += '.zip'
        # if not self.reactive: return zip_filepath

        # assert self.reactive,f'Folder {self.path} is not reactive'

        if os.path.exists(zip_filepath):
            if forced:
                os.remove(zip_filepath)
            else:
                return zip_filepath

        create_zipfile(self.path, zip_filepath)
        return zip_filepath

    @logger.trace(skimpy=True)
    def unzip(self, zip_filepath, create_subdir=True):
        # assert self.reactive,f'Folder {self.path} is not reactive'
        # if save_path_formatter is None: save_path_formatter=lambda x:x
        # zip_path=[zip_filepath,
        #           zip_filepath+'.zip',
        #           self.get_filepath(zip_filepath),
        #           self.get_filepath(zip_filepath,ext='.zip')]
        # zip_path=get_existing_path(zip_path)
        # assert not os.path.isdir(zip_path),f'The folder "{zip_filepath}" cannot be unzipped.'
        if create_subdir:
            zipfilename = os.path.splitext(Path(zip_filepath).name)[0]
            # logger.debug('zipfilename: %s',zipfilename)
            dst_folder_path = self.create(zipfilename).path
        # logger.debug('dst_folder_path: %s',dst_folder_path)
        else:
            dst_folder_path = self.path
        if not self.reactive: return self._inherit_class(dst_folder_path)
        shutil.unpack_archive(zip_filepath, dst_folder_path, 'zip')
        return self._inherit_class(dst_folder_path)

    @logger.trace(skimpy=True)
    def rename(self, new_name):
        new_path = self.parent().get_filepath(new_name)
        # new_path=os.path.join(get_parent_dir(self.path),new_name)
        if self.reactive: os.rename(self.path, new_path)
        self.path = new_path
        return self

    @logger.trace(skimpy=True)
    def move_to(self, dst_path):
        if not isinstance(dst_path, str): dst_path = dst_path.path
        # if self.reactive: shutil.move(self.path,dst_path)
        shutil.move(self.path, dst_path)
        self.path = os.path.join(dst_path, self.name)
        return self

    def parent(self, depth=1):
        """

        :rtype: Folder
        """
        return self._inherit_class(get_parent_dir(self.path, depth=depth))

    @logger.trace(skimpy=True)
    def copy_to(self, dst_folder, new_name='', forced=False):
        """

        :rtype: Folder
        """
        new_name = new_name or self.name
        dst_folder_path = self._inherit_class(dst_folder).get_filepath(new_name)
        
        delete_dst = lambda: shutil.rmtree(dst_folder_path, ignore_errors=True)
        resulter = lambda: self._class(dst_folder_path)

        if os.path.exists(dst_folder_path):
            if forced:
                delete_dst()
            else:
                source_hashsum = dirhash(self.path)
                destination_hashsum = dirhash(dst_folder_path)
                if source_hashsum == destination_hashsum:
                    return resulter()
                else:
                    delete_dst()

        shutil.copytree(self.path, dst_folder_path)
        return resulter()

    def size(self, byte_factor=1e-9):
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(self.path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)
        return total_size * byte_factor

    @property
    def exists(self):
        return self._exists(self.path)

    # def create_iterated(self,start=None,delim='_',empty_ok=False):
    # 	if empty_ok:
    # 		validator_func=lambda existing_path:not self._inherit_class(existing_path).children
    # 	else:
    # 		validator_func=lambda existing_path:existing_path==self.path
    #
    # 	new_iter_folder_path=create_iterated_path(self.path,start,delim,validator_func=validator_func)
    # 	last_part=new_iter_folder_path.split(delim)[-1]
    # 	if last_part.isnumeric() and int(last_part)==(start or 0):
    # 		new_folder=self.rename(new_iter_folder_path)
    # 	else:
    # 		new_folder=self.__class__(new_iter_folder_path)
    # 		if not new_folder==self and not self.children(): self.delete()
    # 	return new_folder

    def __getitem__(self, item):
        item = str(item)
        if os.path.splitext(item)[-1]:
            return self.get_filepath(item)
        else:
            return self.create(item)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.path == other
        else:
            return self.path == other.path

    def __setitem__(self, key, value):
        filename, ext = os.path.splitext(key)
        if isinstance(value, str):
            string_obj = value
        elif type(value) in [list, tuple]:
            string_obj = '\n'.join(map(str, value))
        elif isinstance(value, dict):
            ext = ext or '.yaml'
            if ext in ['.yaml', '.yml']:
                Box(value).to_yaml(filename=self.get_filepath(filename, ext=ext))
                return
            elif ext == '.json':
                Box(value).to_json(filename=self.get_filepath(filename, ext=ext))
                return
            else:
                string_obj = '\n'.join(map(lambda kv: f"{kv[0]}: {kv[1]}", value.items()))

        else:
            string_obj = str(value)
        # def writer(filehandler):
        # 	for k,v in value.items():
        # 		filehandler.writelines([f"{k}: {v}"])

        with open(self.get_filepath(filename, ext=ext or '.txt'), 'w') as fh:
            fh.write(str(string_obj))

    def __repr__(self):
        return f"{self.__class__.__name__}(r'{self.path}')"

    # return self.path

    def __fspath__(self):
        return self.path


def create_iterated_path(iterated_path: str, start=None, delim='_', validator_func=None) -> str:
    if validator_func is None: validator_func = lambda x: False
    base_path, item_name_with_ext = os.path.split(iterated_path)
    item_name, ext = os.path.splitext(item_name_with_ext)

    curr_iter = 0 if start is None else start
    while True:
        new_item_name = delim.join([item_name, f'{curr_iter}'])
        new_itempath = os.path.join(base_path, new_item_name + ext)
        if not os.path.exists(new_itempath) or validator_func(new_itempath): return new_itempath
        curr_iter += 1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lgblkb_tools', 'lgblkb_tools.common', 'lgblkb_tools.db']

package_data = \
{'': ['*']}

install_requires = \
['bumpversion>=0.5.3,<0.6.0',
 'checksumdir>=1.1.7,<2.0.0',
 'colorlog>=4.1.0,<5.0.0',
 'dynaconf>=2.2.2,<3.0.0',
 'geoalchemy2>=0.6.3,<0.7.0',
 'geojson>=2.5.0,<3.0.0',
 'geopandas>=0.7.0,<0.8.0',
 'gitchangelog>=3.0.4,<4.0.0',
 'invoke>=1.4.1,<2.0.0',
 'matplotlib>=3.1.3,<4.0.0',
 'more-itertools>=8.2.0,<9.0.0',
 'numpy>=1.18.1,<2.0.0',
 'opencv-python>=4.2.0,<5.0.0',
 'pandas<1.0.0',
 'pyproj>=2.5.0,<3.0.0',
 'python-box<=4.0.0',
 'python-log-indenter>=0.9,<0.10',
 'python-telegram-bot>=12.4.2,<13.0.0',
 'pyyaml>=5.3,<6.0',
 'requests>=2.23.0,<3.0.0',
 'scipy>=1.4.1,<2.0.0',
 'shapely>=1.7.0,<2.0.0',
 'sqlalchemy>=1.3.13,<2.0.0',
 'wrapt>=1.12.0,<2.0.0']

entry_points = \
{'console_scripts': ['bump = lgblkb_tools.dev_script:version_patch',
                     'export_reqs = lgblkb_tools.dev_script:export_reqs',
                     'publish = lgblkb_tools.dev_script:publish']}

setup_kwargs = {
    'name': 'lgblkb-tools',
    'version': '1.1.12',
    'description': 'Tools to make life easier',
    'long_description': None,
    'author': 'lgblkb',
    'author_email': 'dbakhtiyarov@nu.edu.kz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.9,<4.0.0',
}


setup(**setup_kwargs)
